import pandas as pd

import bionty as bt


def test_efo_readout_curation_ontology_id():
    df = pd.DataFrame(
        index=[
            "EFO:0000001",
            "EFO:1002050",
            "EFO:1002047",
            "EFO:0000005",
            "This readout does not exist",
        ]
    )
    curated_df = bt.Readout(database="efo", version="3.48.0").curate(df)

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)


def test_efo_readout_curation_name():
    df = pd.DataFrame(
        index=[
            "CS57511",
            "CS57520",
            "CS57515",
            "experimental factor",
            "This readout does not exist",
        ]
    )
    curated_df = bt.Readout(id="name", database="efo", version="3.48.0").curate(df)

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)



from typing import Literal, Optional

import pandas as pd
from cached_property import cached_property

from .._entity import Entity
from .._normalize import GENE_COLUMNS, NormalizeColumns
from .._settings import s3_bionty_assets
from ._shared_docstrings import _doc_params, doc_entites

ALIAS_DICT = {"symbol": "synonyms"}


@_doc_params(doc_entities=doc_entites)
class Gene(Entity):
    """Gene.

    1. Ensembl
    Edits of terms are coordinated and reviewed on:
    https://github.com/geneontology/

    The default indexer is `ensembl_gene_id`

    Args:
        {doc_entities}

    Notes:
        Biotypes: https://www.ensembl.org/info/genome/genebuild/biotypes.html
        Gene Naming: https://www.ensembl.org/info/genome/genebuild/gene_names.html
    """

    def __init__(
        self,
        species: str = "human",
        id: Optional[str] = None,
        database: Optional[Literal["ensembl"]] = None,
        version: Optional[str] = None,
    ):
        super().__init__(id=id, database=database, version=version, species=species)
        self._lookup_col = "symbol"

    @cached_property
    def df(self):
        """DataFrame.

        See ingestion: https://lamin.ai/docs/bionty-assets/ingest/ensembl-gene
        """
        cloudpath = s3_bionty_assets(self._cloud_file_path)
        self._filepath = cloudpath.fspath

        df = pd.read_parquet(self._filepath)
        NormalizeColumns.gene(df, species=self.species)
        if not df.index.is_numeric():
            df = df.reset_index().copy()
        df = df[~df[self._id].isnull()]
        return df.set_index(self._id)

    def curate(  # type: ignore
        self, df: pd.DataFrame, column: str = None
    ) -> pd.DataFrame:
        """Curate index of passed DataFrame to conform with default identifier.

        - If `column` is `None`, checks the existing index for compliance with
          the default identifier.
        - If `column` denotes an entity identifier, tries to map that identifier
          to the default identifier.

        Returns the DataFrame with the curated index and a boolean `__curated__`
        column that indicates compliance with the default identifier.

        In addition to the .curate() in base class, this also performs alias mapping.
        """
        agg_col = ALIAS_DICT.get(self._id)
        df = df.copy()

        # if the query column name does not match any columns in the self.df
        # Bionty assume the query column and the self._id_field uses the same type of
        # identifier
        orig_column = column
        if column is not None and column not in self.df.columns:
            # normalize the identifier column
            column_norm = GENE_COLUMNS.get(column)
            if column_norm in df.columns:
                raise ValueError("{column_norm} column already exist!")
            else:
                column = self._id if column_norm is None else column_norm
                df.rename(columns={orig_column: column}, inplace=True)
            agg_col = ALIAS_DICT.get(column)

        return (
            super()
            ._curate(df=df, column=column, agg_col=agg_col)
            .rename(columns={column: orig_column})
        )



import os
from pathlib import Path

from rich.console import Console
from rich.table import Table

from bionty._settings import settings
from bionty.dev._io import load_yaml

console = Console()

ROOT_DIR = Path(__file__).parent.resolve()


def display_available_versions() -> None:  # pragma: no cover
    """Displays all available entities and versions in a Rich table."""
    VERSIONS_FILE_PATH = Path(f"{settings.versionsdir}/local.yaml").resolve()
    versions = load_yaml(VERSIONS_FILE_PATH.resolve())

    table = Table(title="Available versions")

    table.add_column("Ontology", justify="right", style="cyan", no_wrap=True)
    table.add_column("URL", justify="right", style="cyan", no_wrap=True)
    table.add_column("Bionty Entity", justify="right", style="cyan", no_wrap=True)
    table.add_column("Database key", justify="right", style="cyan", no_wrap=True)
    table.add_column("All versions", justify="right", style="cyan", no_wrap=True)

    for entity, db_to_version in versions.items():
        for db, _to_versions_url in db_to_version.items():
            versions = ""
            _ontology_name = _to_versions_url["name"]
            _ontology_url = _to_versions_url["website"]
            for version_str, url in _to_versions_url["versions"].items():
                versions += str(version_str) + "\n"
            table.add_row(_ontology_name, _ontology_url, entity, db, versions)

    console.print(table)


def display_active_versions() -> None:  # pragma: no cover
    """Displays all currently set as default entities and versions in a Rich table."""
    version_table = (
        "._lndb.yaml" if os.getenv("LAMINDB_INSTANCE_LOADED") == 1 else "._current.yaml"
    )

    VERSIONS_FILE_PATH = Path(f"{ROOT_DIR}/versions/{version_table}").resolve()
    versions = load_yaml(VERSIONS_FILE_PATH.resolve())

    table = Table(title=f"Currently used versions in {version_table}")

    table.add_column("Entity", justify="right", style="cyan", no_wrap=True)
    table.add_column("Database", justify="right", style="cyan", no_wrap=True)
    table.add_column("Version", justify="right", style="cyan", no_wrap=True)

    for entity, db_to_version in versions.items():
        for db, version in db_to_version.items():
            table.add_row(entity, db, str(version))

    console.print(table)



import pandas as pd

import bionty as bt


def test_uniprot_protein_curation_ontology_id():
    df = pd.DataFrame(
        index=[
            "A0A024QZ08",
            "X6RLV5",
            "X6RM24",
            "A0A024QZQ1",
            "This cell line does not exist",
        ]
    )
    curated_df = bt.Protein(database="uniprot", version="2022-04").curate(df)

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)


def test_uniprot_protein_curation_name():
    df = pd.DataFrame(
        index=[
            "isoform CRA_c",
            "Battenin",
            "Probable ATP-dependent RNA helicase DDX5",
            "isoform CRA_a",
            "This cell line does not exist",
        ]
    )
    curated_df = bt.Protein(id="name", database="uniprot", version="2022-04").curate(df)

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)



import requests  # type: ignore


def ontology_info(namespace: str) -> dict:
    """Get the latest information of an ontology from OLS.

    Args:
        namespace: ontology namespace (e.g: cl, efo, go)
    """
    headers = {"Accept": "application/json"}
    r = requests.get(
        f"http://www.ebi.ac.uk/ols/api/ontologies/{namespace.lower()}", headers=headers
    )

    if r.status_code != 200:
        raise RuntimeError(f"No information found about ontology {namespace}.")

    return r.json().get("config")



from typing import Literal, Optional

import pandas as pd
from cached_property import cached_property

from .._entity import Entity
from .._normalize import NormalizeColumns
from .._settings import s3_bionty_assets
from ._shared_docstrings import _doc_params, doc_entites


@_doc_params(doc_entities=doc_entites)
class Protein(Entity):
    """Protein.

    1. Uniprot
    Edits of terms are coordinated and reviewed on:
    https://www.uniprot.org/

    Args:
        {doc_entities}
    """

    def __init__(
        self,
        species: str = "human",
        id: Optional[str] = None,
        database: Optional[Literal["uniprot"]] = None,
        version: Optional[str] = None,
    ) -> None:
        super().__init__(id=id, database=database, version=version, species=species)
        self._id_field = "uniprotkb_id" if id is None else id

    @cached_property
    def df(self) -> pd.DataFrame:
        """DataFrame.

        See ingestion: https://lamin.ai/docs/bionty-assets/ingest/uniprot-protein
        """
        cloudpath = s3_bionty_assets(self._cloud_file_path)
        self._filepath = cloudpath.fspath

        df = pd.read_parquet(self._filepath)
        NormalizeColumns.protein(df)
        _get_shortest_name(
            df, "synonyms"
        )  # Take the shortest name in protein names list as name
        if not df.index.is_numeric():
            df = df.reset_index().copy()
        df = df[~df[self._id_field].isnull()]

        return df.set_index(self._id_field)


def _get_shortest_name(df: pd.DataFrame, column: str, new_column="name"):
    """Get a single shortest name from a column of lists."""
    name_list = []
    names_list = []
    for lst in df[column]:
        lst = lst.replace(", ", "|")
        names_list.append(lst)

        def shortest_name(lst: list):
            return min(lst, key=len)

        names = lst.split("|")
        no_space_names = [i for i in names if " " not in i]
        if len(no_space_names) == 0:
            name = shortest_name(names)
        else:
            name = shortest_name(no_space_names)
        name_list.append(name)

    df[new_column] = name_list
    df[column] = names_list



import pytest

from bionty import CellMarker, Gene, Protein
from bionty._entity import Entity, _camel_to_snake


def test_entity_table():
    with pytest.raises(AttributeError):
        entity_table = Entity(database="mondo")
        assert entity_table.df is not None
        assert entity_table.entity == "entity_table"
        assert _camel_to_snake("EntityTable") == "entity_table"


def species_not_implemented():
    with pytest.raises(NotImplementedError):
        CellMarker(species="mouse")
    with pytest.raises(NotImplementedError):
        Protein(species="cat")
    with pytest.raises(NotImplementedError):
        Gene(species="cat")



import pandas as pd

import bionty as bt


def test_mondo_disease_curation_ontology_id():
    df = pd.DataFrame(
        index=[
            "UBERON:8440004",
            "UPHENO:0001001",
            "UBERON:8420000",
            "http://identifiers.org/hgnc/10001",
            "This disease does not exist",
        ]
    )
    curated_df = bt.Disease(database="mondo", version="2023-02-06").curate(df)

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)


def test_mondo_disease_curation_name():
    df = pd.DataFrame(
        index=[
            "laminar subdivision of the cortex",
            "hair of scalp",
            "RHAG",
            "GRK1",
            "This disease does not exist",
        ]
    )
    curated_df = bt.Disease(id="name", database="mondo", version="2023-02-06").curate(
        df
    )

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)


def test_doid_disease_curation_ontology_id():
    df = pd.DataFrame(
        index=[
            "DOID:0001816",
            "DOID:0002116",
            "DOID:5547",
            "DOID:5551",
            "This disease does not exist",
        ]
    )
    curated_df = bt.Disease(database="doid", version="2023-01-30").curate(df)

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)


def test_doid_disease_curation_name():
    df = pd.DataFrame(
        index=[
            "aspirin allergy",
            "foramen magnum meningioma",
            "skull base meningioma",
            "choriocarcinoma of the testis",
            "This disease does not exist",
        ]
    )
    curated_df = bt.Disease(id="name", database="doid", version="2023-01-30").curate(df)

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)



import pandas as pd

import bionty as bt


def test_uberon_tissue_curation_ontology_id():
    df = pd.DataFrame(
        index=[
            "UBERON:0000000",
            "UBERON:0000005",
            "UBERON:8600001",
            "UBERON:8600002",
            "This tissue does not exist",
        ]
    )
    curated_df = bt.Tissue(database="uberon", version="2023-02-14").curate(df)

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)


def test_uberon_tissue_curation_name():
    df = pd.DataFrame(
        index=[
            "nose",
            "chemosensory organ",
            "epithelium of lobular bronchiole",
            "smooth muscle tissue of lobular bronchiole",
            "This tissue does not exist",
        ]
    )
    curated_df = bt.Tissue(id="name", database="uberon", version="2023-02-14").curate(
        df
    )

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)



import pandas as pd

import bionty as bt


def test_uniprot_cellmarker_curation_ontology_id():
    df = pd.DataFrame(
        index=["CCR7", "CD69", "CD8A", "CD45RA", "This protein does not exist"]
    )

    curated_df = bt.CellMarker(database="cellmarker", version="2.0").curate(df)

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)


def test_uniprot_cellmarker_curation_name():
    df = pd.DataFrame(
        index=[
            "1-MAR",
            "160 KDA NEUROFILAMENT MEDIUM",
            "VON WILLEBRAND FACTOR",
            "TYPE IV COLLAGEN",
            "This protein does not exist",
        ]
    )
    curated_df = bt.CellMarker(id="name", database="cellmarker", version="2.0").curate(
        df
    )

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)



from typing import Literal, Optional

import pandas as pd
from cached_property import cached_property

from bionty.entities._shared_docstrings import _doc_params, species_removed

from .._entity import Entity
from .._ontology import Ontology
from .._settings import check_datasetdir_exists, settings

EFO_DF_D3 = "https://bionty-assets.s3.amazonaws.com/efo_df.json"


@_doc_params(doc_entities=species_removed)
class Readout(Entity):
    """Experimental Factor.

    1. Experimental Factor Ontology
    Edits of terms are coordinated and reviewed on:
    https://www.ebi.ac.uk/ols/ontologies/efo

    Args:
        {doc_entities}

    Also see: `bionty.Entity <https://lamin.ai/docs/bionty/bionty.entity>`__
    """

    def __init__(
        self,
        id: str = "ontology_id",
        database: Optional[Literal["efo"]] = None,
        version: Optional[str] = None,
    ) -> None:
        super().__init__(id=id, database=database, version=version)
        self._filepath = settings.datasetdir / "efo_df.json"
        self._readout_terms = {
            "assay": "OBI:0000070",
            "assay_by_molecule": "EFO:0002772",
            "assay_by_instrument": "EFO:0002773",
            "assay_by_sequencer": "EFO:0003740",
            "measurement": "EFO:0001444",
        }

    @cached_property
    def df(self) -> pd.DataFrame:
        """DataFrame."""
        if not self._filepath.exists():
            self._download_df()
        df = pd.read_json(self._filepath)
        df.index.name = "ontology_id"

        if self._id != "ontology_id":
            return df.reset_index().set_index(self._id)
        else:
            return df

    @cached_property
    def ontology(self) -> Ontology:  # type:ignore
        """EFO."""
        localpath = self._url_download(self._url)

        return Ontology(handle=localpath, prefix="http://www.ebi.ac.uk/efo/")

    @cached_property
    def assay(self) -> list:
        """Assays."""
        # OBI:0000070
        return self.ontology._list_subclasses(self._readout_terms["assay"])

    @cached_property
    def assay_by_molecule(self) -> list:
        """Assays by molecule."""
        # EFO:0002772
        return self.ontology._list_subclasses(self._readout_terms["assay_by_molecule"])

    @cached_property
    def assay_by_instrument(self) -> list:
        """Assays by instrument."""
        # EFO:0002773
        return self.ontology._list_subclasses(
            self._readout_terms["assay_by_instrument"]
        )

    @cached_property
    def assay_by_sequencer(self) -> list:
        """Assay by sequencer."""
        # EFO:0003740
        return self.ontology._list_subclasses(self._readout_terms["assay_by_sequencer"])

    @cached_property
    def measurement(self) -> list:
        """Measurement."""
        # EFO:0001444
        return self.ontology._list_subclasses(self._readout_terms["measurement"])

    @check_datasetdir_exists
    def _download_df(self) -> None:
        from urllib.request import urlretrieve

        urlretrieve(
            EFO_DF_D3,
            self._filepath,
        )

    def get(self, term_id: str) -> dict:
        """Parse readout attributes from EFO."""

        def _list_to_str(lst: list):
            if len(lst) == 0:
                return None
            elif len(lst) == 1:
                return lst[0].name  # type: ignore
            else:
                return ";".join([i.name for i in lst])

        term = self.ontology.get_term(term_id)
        superclasses = term.superclasses()

        # get the molecule term
        molecules = [i for i in self.assay_by_molecule if i in superclasses]
        # get the instrument term
        instruments = [i for i in self.assay_by_sequencer if i in superclasses]
        if len(instruments) == 0:
            instruments = [i for i in self.assay_by_instrument if i in superclasses]
        # get the measurement for non-molecular readouts
        measurements = [i for i in self.measurement if i in superclasses]

        readout = {
            "ontology_id": term_id,
            "name": term.name,
            "molecule": _list_to_str(molecules),
            "instrument": _list_to_str(instruments),
            "measurement": _list_to_str(measurements),
        }

        return readout



import os
import urllib.request
from pathlib import Path
from typing import Union
from urllib.error import HTTPError, URLError

import pandas as pd
import yaml  # type:ignore


def load_yaml(filename: Union[str, Path]):  # pragma: no cover
    with open(filename, "r") as f:
        return yaml.safe_load(f)


VERSIONS_FILE_PATH = Path(f"{os.getcwd()}/bionty/versions/versions.yaml")

versions = load_yaml(VERSIONS_FILE_PATH.resolve())

# We currently assume that the versions.yaml file has the URLs as the final values
flattened_df = pd.json_normalize(versions, sep="_")
flattened_dict = flattened_df.to_dict(orient="records")[0]

failed_urls = []
for url in flattened_dict.values():
    if url.startswith("https://"):
        try:
            assert urllib.request.urlopen(url, timeout=100).getcode() == 200
        except (AssertionError, ValueError, HTTPError, URLError) as e:
            failed_urls.append([url, e])

if len(failed_urls) != 0:
    for fail in failed_urls:
        print(fail)
    raise AssertionError(f"{len(failed_urls)} URLs failed.")



from typing import Literal, Optional

from .._entity import Entity
from ._shared_docstrings import _doc_params, doc_entites


@_doc_params(doc_entities=doc_entites)
class Phenotype(Entity):
    """Phenotype.

    1. Human Phenotype Ontology
    Edits of terms are coordinated and reviewed on:
    https://hpo.jax.org/app/

    Args:
        {doc_entities}
    """

    def __init__(
        self,
        species: str = "human",
        id: str = "ontology_id",
        database: Optional[Literal["hp"]] = None,
        version: Optional[str] = None,
    ) -> None:
        super().__init__(
            id=id,
            database=database,
            version=version,
            species=species,
        )
